#include<stdlib.h>
#include<stdio.h>
#include <math.h>

#include<omp.h>
#include <CL/cl.h>

#include"uth_ocl_intf.h"

#include"./vecadd_host.h"

#define time_measurments

#ifdef time_measurments
#include"uth_time_intf.h"
  static double t_begin, t_end, t_total;
#endif

int create_memory_objects(
			  int platform_index,
			  int device_index,
			  int array_size,
			  float* a,
			  float* b,
			  cl_mem* memObjects,
			  int monitor
			  )
{
  
  cl_int retval;
  
  // choose the context
  cl_context context = utr_ocl_select_context(platform_index, device_index);  
  
  if(monitor>UTC_BASIC_INFO){
    printf("\n\t\t1. creating memory objects\n");
  }
  
#ifdef time_measurments
  t_begin = time_clock();
#endif
  
  // version 1: creating memory objects with CL_MEM_COPY_HOST_PTR option
  //memObjects[0] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
  //				 sizeof(float) * array_size, a, NULL);
  //  memObjects[1] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
  //				 sizeof(float) * array_size, b, NULL);
  
  // version 2: just creating memory objects with the content copied explicitly
  memObjects[0] = clCreateBuffer(context, CL_MEM_READ_ONLY ,
  				 sizeof(float) * array_size, NULL, NULL);
  memObjects[1] = clCreateBuffer(context, CL_MEM_READ_ONLY ,
  				 sizeof(float) * array_size, NULL, NULL);
  memObjects[2] = clCreateBuffer(context, CL_MEM_READ_WRITE,
  				 sizeof(float) * array_size, NULL, NULL);
  
  // Creating pinned memory buffers
  //#define SCALAR float
  /* cl_mem cmPinnedBufIn1 = NULL; // input buffer mapped to host input buffer */
  /* cmPinnedBufIn1 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, */
  /* 				  sizeof(float) * array_size, NULL, NULL); */
  
  /* SCALAR* a = NULL; // host input buffer (standard array) */
  /* a = (SCALAR*)clEnqueueMapBuffer(command_queue, cmPinnedBufIn1, CL_TRUE, */
  /* 				  CL_MAP_WRITE, 0, sizeof(float) * array_size, 0, */
  /* 				  NULL, NULL, NULL); */
  
  /* memObjects[0] = NULL; // device input buffer allocated in card's memory */
  /* memObjects[0] = clCreateBuffer(context, CL_MEM_READ_ONLY, */
  /* 			       sizeof(float) * array_size, NULL, NULL); */
  
  
  /* cl_mem cmPinnedBufIn2 = NULL; // input buffer mapped to host input buffer */
  /* cmPinnedBufIn2 = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, */
  /* 				  sizeof(float) * array_size, NULL, NULL); */
  /* SCALAR* b = NULL; // host input buffer (standard array) */
  /* b = (SCALAR*)clEnqueueMapBuffer(command_queue, cmPinnedBufIn2, CL_TRUE, */
  /* 				  CL_MAP_WRITE, 0, sizeof(float) * array_size, 0, */
  /* 				  NULL, NULL, NULL); */
  
  /* memObjects[1] = NULL; // device input buffer allocated in card's memory */
  /* memObjects[1] = clCreateBuffer(context, CL_MEM_READ_ONLY, */
  /* 			       sizeof(float) * array_size, NULL, NULL); */
  
  
  /* cl_mem cmPinnedBufOut = NULL; // output buffer mapped to host output buffer */
  /* cmPinnedBufOut = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR, */
  /* 				  sizeof(float) * array_size, NULL, NULL); */
  
  /* SCALAR* result = NULL; // host output buffer (standard array) */
  /* result = (SCALAR*)clEnqueueMapBuffer(command_queue, cmPinnedBufOut, CL_TRUE, */
  /* 				       CL_MAP_READ, 0, sizeof(float) * array_size, 0, */
  /* 				       NULL, NULL, NULL); */
  
  /* memObjects[2] = NULL; // device output buffer allocated in card's memory */
  /* memObjects[2] = clCreateBuffer(context, CL_MEM_WRITE_ONLY, */
  /* 			       sizeof(float) * array_size, NULL, NULL); */
  
  if (memObjects[0] == NULL || memObjects[1] == NULL || memObjects[2] == NULL) {
    printf("Error creating memory objects.\n");
    exit(-1);
  }
  
#ifdef time_measurments
  t_end = time_clock();
  printf("EXECUTION TIME: creating memory objects: %lf\n", t_end-t_begin);
#endif
  

  // Copy memory objects to device memory. 
  
  if(monitor>UTC_BASIC_INFO){
    printf("\t\t2. copying memory objects being input data\n");
  }
  
  // choose the command queue
  cl_command_queue command_queue = 
    utr_ocl_select_command_queue(platform_index, device_index);  
  
#ifdef time_measurments
  clFinish(command_queue);
  t_begin = time_clock();
#endif
  
  cl_event event_write_a;
  retval = clEnqueueWriteBuffer(
				command_queue, 
				memObjects[0], 
				CL_FALSE, 
				0, 
				sizeof(float) * array_size,
				a, 
				0, 
				NULL, 
				&event_write_a);
  
  clWaitForEvents(1, &event_write_a);
  
  cl_event event_write_b;
  // Use clEnqueueWriteBuffer() to write input array B to 
  // the device buffer bufferB
  retval = clEnqueueWriteBuffer(
				command_queue, 
				memObjects[1], 
				CL_FALSE, 
				0, 
				sizeof(float) * array_size,    
				b,
				0, 
				NULL, 
				&event_write_b);

  clWaitForEvents(1, &event_write_b);
  
#ifdef time_measurments
  clFinish(command_queue);
  t_end = time_clock();
  // Get kernel profiling info 
  cl_ulong startTime_a;
  cl_ulong endTime_a;
  clGetEventProfilingInfo(event_write_a,
			  CL_PROFILING_COMMAND_START,
			  sizeof(cl_ulong),
			  &startTime_a,
			  0);
  clGetEventProfilingInfo(event_write_a,
			  CL_PROFILING_COMMAND_END,
			  sizeof(cl_ulong),
			  &endTime_a,
			  0);
  cl_ulong startTime_b;
  cl_ulong endTime_b;
  clGetEventProfilingInfo(event_write_b,
			  CL_PROFILING_COMMAND_START,
			  sizeof(cl_ulong),
			  &startTime_b,
			  0);
  clGetEventProfilingInfo(event_write_b,
			  CL_PROFILING_COMMAND_END,
			  sizeof(cl_ulong),
			  &endTime_b,
			  0);

  printf("EXECUTION TIME: copying memory objects: %lf (profiler: %lf)\n", 
	 t_end-t_begin, ((double)endTime_a - (double)startTime_a
			 + (double)endTime_a - (double)startTime_a)*1.0e-9);
  printf("\tsize of data %lf GB, transfer rate %lf GB/s\n", 
	 2*sizeof(float) * array_size * 1.e-9,
	 2*sizeof(float) * array_size * 1.e-9 / (t_end-t_begin));
#endif
  
  
  return(0);
}


int check_result(
		 int platform_index,
		 int device_index,
		 int array_size,
		 float* result,
		 cl_mem* memObjects,
		 int monitor
		 )
{
  int i,j,k;
  
  // choose the context
  cl_context context = utr_ocl_select_context(platform_index, device_index);  
  
  cl_command_queue command_queue = 
    utr_ocl_select_command_queue(platform_index, device_index);  
  
  if(monitor>UTC_BASIC_INFO){
    printf("\n\t\t5. copying memory objects being output data\n");
  }
  
#ifdef time_measurments
  clFinish(command_queue);
  t_begin = time_clock();
#endif
  
  // clear array of results
  for (i = 0; i < array_size; i++) result[i]=0.0;
  
  // Read the output buffer back to the Host
  cl_int retval = clEnqueueReadBuffer(command_queue, memObjects[2], CL_TRUE,
				      0, array_size * sizeof(float), result,
				      0, NULL, NULL);
  
#ifdef time_measurments_1
  clFinish(command_queue);
  t_end = time_clock();
  printf("EXECUTION TIME: copying output memory objects: %lf\n", t_end-t_begin);
  printf("\tsize of data %lf GB, transfer rate %lf GB/s\n",
	 sizeof(float) * array_size * 1.e-9,
	 sizeof(float) * array_size * 1.e-9 / (t_end-t_begin));
#endif
  
  if (retval != CL_SUCCESS){
    printf("Failed to read result buffer.\n");
    //Cleanup(context, command_queue, program, kernel, memObjects);
    exit(-1);
  }
  
  if(monitor>UTC_BASIC_INFO){
    printf("\t\t6. verifying results\n");
  }
  
  // Verify the output
  int result_OK = 1;
  for(i = 0; i < array_size; i++) {
    //printf("%6.0f", result[i]);
    if(fabs(result[i]-(float)(i+2.0*i))/result[i]>1.0e-6){
      result_OK = 0;
      break;
    }
  }
  if(result_OK) {
    printf("Output is correct\n");
  } else {
    printf("Output is incorrect\n");
    //exit(-1);
  }
  
  return(0);  
}

	       
