#include<stdlib.h>
#include<stdio.h>
#include <math.h>

#include<omp.h>
#include <CL/cl.h>

#include"uth_ocl_intf.h"

#include"./vecadd_host.h"

// HARDWARE DEPENDENT PARAMETERS
//#define NR_COMP_UNITS 
//#define NR_CORES_PER_CU 
//#define NR_WORK_GROUPS_PER_CU 
//#define WORK_GROUP_SIZE 
//#define NR_ENTRIES_PER_THREAD 
#include"./hardware_params.h"

#define NR_CORES (NR_CORES_PER_CU*NR_COMP_UNITS)
#define NR_WORK_GROUPS (NR_COMP_UNITS*NR_WORK_GROUPS_PER_CU)
#define NR_THREADS (NR_WORK_GROUPS*WORK_GROUP_SIZE)
#define ARRAY_SIZE (NR_THREADS*NR_ENTRIES_PER_THREAD)

#define time_measurments

#ifdef time_measurments
#include"uth_time_intf.h"
  static double t_begin, t_end, t_total;
#endif

int Platform_ID = 0;

int main(int argc, char** argv)
{
  
  int i;
  
  // for all operations indicate explicit info messages
  int monitor = UTC_BASIC_INFO + 1;
  
  
  /*----------------STANDARD CPU EXECUTION PHASE----------------------*/
  
  if(monitor>UTC_BASIC_INFO){
    printf("\n------------Starting OpenMP execution phase----------------\n");
    printf("\nArrays size: %d (%d MB)\n", ARRAY_SIZE, ARRAY_SIZE*sizeof(float)/1024/1024);
  }
  
  // create standard (host) data structures and GPU memory buffers
  // used to store the arguments to the kernel
  float *result;
  result = (float *)malloc(ARRAY_SIZE*sizeof(float));
  float *a;
  a = (float *)malloc(ARRAY_SIZE*sizeof(float));
  float *b;
  b = (float *)malloc(ARRAY_SIZE*sizeof(float));
  
  // initialize input arrays
  for (i = 0; i < ARRAY_SIZE; i++)
    {
      a[i] = (float)i;
      b[i] = (float)(i * 2);
    }
  
  // clear array of results
  for (i = 0; i < ARRAY_SIZE; i++) result[i]=0.0;
  
#ifdef time_measurments
  printf("\nExecuting standard loop on one core\n"); 
  t_begin = time_clock();
#endif
  
  // perform standard calculations
  for(i=0;i<ARRAY_SIZE;i++){
    result[i] = a[i] + b[i];
  }
  
#ifdef time_measurments
  t_end = time_clock();
  printf("EXECUTION TIME: executing standard loop: %lf\n", t_end-t_begin);
  printf("\tNumber of operations %d, performance %lf GFlops\n",
	 ARRAY_SIZE, ARRAY_SIZE / (t_end-t_begin) * 1e-9);
  printf("\tGBytes transferred to processor %lf, speed %lf GB/s\n",
	 3*ARRAY_SIZE*sizeof(float)*1e-9, 
	 3*ARRAY_SIZE*sizeof(float)/(t_end-t_begin)*1e-9);
#endif
  
#pragma omp parallel 
  {
    
#pragma omp for
    for(i=0;i<ARRAY_SIZE;i++){
      result[i] = a[i] + b[i];
    }
    
    // clear array of results
#pragma omp for
    for (i = 0; i < ARRAY_SIZE; i++) result[i]=0.0;
    
#ifdef time_measurments
    printf("\nExecuting OpenMP loop\n"); 
    if(omp_get_thread_num()==0) t_begin = time_clock();
#endif
    
#pragma omp for
    for(i=0;i<ARRAY_SIZE;i++){
      result[i] = a[i] + b[i];
    }
    
#ifdef time_measurments
    if(omp_get_thread_num()==0) t_end = time_clock();
#endif
    
  }
  
#ifdef time_measurments
  printf("EXECUTION TIME: executing OpenMP loop: %lf\n", t_end-t_begin);
  printf("\tNumber of operations %d, performance %lf GFlops\n",
	 ARRAY_SIZE, ARRAY_SIZE / (t_end-t_begin) * 1e-9);
  printf("\tGBytes transferred to processor %lf, speed %lf GB/s\n",
	 3*ARRAY_SIZE*sizeof(float)*1e-9, 
	 3*ARRAY_SIZE*sizeof(float)/(t_end-t_begin)*1e-9);
#endif
  
  /*----------------INITIALIZATION PHASE----------------------*/
  
#ifdef time_measurments
  t_begin = time_clock();
#endif
  
  // Create OpenCL contexts on all available platforms
  // contexts are stored in table with indices: 0-CPU, 1-GPU, 2-ACCELERATOR
  // table entry is NULL if there is no such context for a given platform
  int number_of_platforms = utr_ocl_create_contexts(UTC_OCL_ALL_PLATFORMS, monitor);
  
  if(number_of_platforms>1){
    printf("\nMore than one platform in a system! Check whether it's OK with the code and read the proper ID:\n");
    scanf("%d",  &Platform_ID);
  }
  
#ifdef time_measurments
  t_end = time_clock();
  printf("EXECUTION TIME: creating contexts: %lf\n", t_end-t_begin);
#endif
  
#ifdef time_measurments
  t_begin = time_clock();
#endif
  
  // for the selected platform 
  int platform_index = Platform_ID;
  // create command queues on all devices 
  utr_ocl_create_command_queues(platform_index, 
				UTC_OCL_ALL_DEVICES, monitor);
  
#ifdef time_measurments
  t_end = time_clock();
  printf("EXECUTION TIME: creating command queues: %lf\n", t_end-t_begin);
#endif
  
  //cl_platform_id platform = utv_ocl_struct.list_of_platforms[0].id;
  utt_ocl_platform_struct platform_struct = utv_ocl_struct.list_of_platforms[platform_index];
  

  /*----------------EXECUTION PHASE----------------------*/
  
  
  if(monitor>UTC_BASIC_INFO){
    printf("\n------------Starting OpenCL execution phase----------------\n");
    printf("\nArrays size: %d (%d MB)\n", ARRAY_SIZE, ARRAY_SIZE*sizeof(float)/1024/1024);
  }
  
  
  // in a loop over devices (or for a selected device)
  int idev; 
  //for(idev=0; 
  //    idev<platform_struct.number_of_devices;
  //    idev++){
  
  // int device_type = .....
  // choose device_index
  // int device_index = utr_ocl_select_device(platform_index, device_type);
  
  if(platform_struct.number_of_devices>1){
    printf("\n\nSelect device index (for platform %d): ", platform_index); 
    scanf("%d", &idev);
  }
  else{
    idev = 0;
  }
  
  int device_index = idev;
  int device_type = utr_ocl_device_type(platform_index, device_index); 
  
  if(device_index>0 && device_type==utr_ocl_device_type(platform_index, device_index-1)){
    printf("Wrong device %d selected. Exiting.\n", device_index);
    exit(-1);
  } 
  
  //************ MANAGING MEMORY OBJECTS (INPUT AND OUTPUT DATA FOR KERNEL) ***********//
  
  // Create memory objects that will be used as arguments to kernel. 
  cl_mem memObjects[3] = { 0, 0, 0 };
  cl_int retval;
    
  int array_size = ARRAY_SIZE;
  create_memory_objects(platform_index, device_index,
			array_size, a, b, memObjects, monitor);
  
  
  //***** THE END OF: MANAGING MEMORY OBJECTS (INPUT AND OUTPUT DATA FOR KERNEL) *****//
  
  int kernel_index;
  do {
    
    
    /* define a menu */
    
    printf("\n\nSelect kernel index (>100 to exit)\n"); 
    printf("(kernel x in file vecadd_x...cl, kernels 0,1 only for GPU): ");
    
    scanf("%d",&kernel_index);
    
    if(device_type == UTC_OCL_DEVICE_CPU){
      
      if(!utr_ocl_CPU_context_exists(platform_index)){
	printf("Device type CPU but no CPU context exists. Exiting.\n");
	//Cleanup(context, command_queue, program, kernel, memObjects);
	exit(-1);
      }
      
            
      if(kernel_index==2){
	
	execute_kernel_2_CPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
	
      }
      
      if(kernel_index==3){
	
	execute_kernel_3_CPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
		
      }

    }
    else if(device_type == UTC_OCL_DEVICE_GPU){

      if(!utr_ocl_GPU_context_exists(platform_index)){
	printf("Device type GPU but no GPU context exists. Exiting.\n");
	//Cleanup(context, command_queue, program, kernel, memObjects);
	exit(-1);
      }

      if(kernel_index==0){
	
	execute_kernel_0_GPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
	
      }
      if(kernel_index==1){
	
	execute_kernel_1_GPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
	
      }
      if(kernel_index==2){
	
	execute_kernel_2_GPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
	
      }
      if(kernel_index==3){
	
	execute_kernel_3_GPU(platform_index, device_index, kernel_index, 
			     array_size, memObjects, monitor);
	
	check_result(platform_index, device_index,
		     array_size, result, memObjects, monitor);
	
      }


    }
    
    
  } while(kernel_index<100);
  
  //} // the end of loop over devices
  
  for (i = 0; i < 3; i++)
    {
      if (memObjects[i] != 0)
	clReleaseMemObject(memObjects[i]);
    }
  
  /*----------------FINALIZATION PHASE----------------------*/
  
  utr_ocl_cleanup();
  
  return 0;
}

